import math
import itertools

import numpy as np

from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit import Parameter
from qiskit.quantum_info import Operator
from qiskit.circuit.library import RZGate

from .constrained_mixer import Constrained

from qaoa.util import dicke_state


class Grover(Constrained):
    def __init__(self) -> None:
        self.__check_params()
        self.k = self.params["k"]
        self.n = self.params["N_qubits"]

        self.U_s = dicke_state(self.n, self.k)
        self.U_s_dagger = self.U_s.inverse()

    # def set_initial_state(self, circuit, qubit_register):
    #    circuit.compose(self.U_s, qubit_register, inplace=True)

    def create_circuit(self):
        q = QuantumRegister(self.params["N_qubits"])
        self.circuit = QuantumCircuit(q)
        # self.best_mixer_terms, self.logical_X_operators = self.__XYMixerTerms()

        Beta = Parameter("x_beta")
        rz = RZGate(Beta).control(self.n - 1)

        self.circuit.compose(self.U_s_dagger, q, inplace=True)
        self.circuit.x(range(self.n))
        self.circuit.append(rz, self.circuit.qubits)
        self.circuit.x(range(self.n))
        self.circuit.compose(self.U_s, self.circuit.qubits, inplace=True)

    def compute_feasible_subspace(self):
        print("Its now computing the feasible subspace")
        self.B.clear()
        for combination in itertools.combinations(
            range(self.params["N_qubits"]), self.k
        ):
            current_state = ["0"] * self.params["N_qubits"]
            for index in combination:
                current_state[index] = "1"
            self.B.append("".join(current_state))

    def isFeasible(self, string):
        x = self.__str2np(string)
        constraint = np.sum(x) - self.k
        return math.isclose(constraint, 0, abs_tol=1e-7)

    def __check_params(self):
        for key in ["k", "N_qubits"]:
            assert key in self.params, "Problem or params need to specify " + key



import numpy as np
from sympy.physics.quantum import TensorProduct
from sympy.physics.paulialgebra import Pauli, evaluate_pauli_product
from OLD.binsymbols import binsymbols
from sympy import *
from sympy.core.numbers import One as sympyOne
from openquantumcomputing.utilities import decompose
import itertools
import math

X = Pauli(1)
Y = Pauli(2)
Z = Pauli(3)
I = 1


def hamming(s1, s2):
    return sum(c1 != c2 for c1, c2 in zip(s1, s2))


def get_T(n, mode, d=1, i=None, j=None, shift=1, oddeven="both"):
    if mode == "leftright":
        T = np.zeros((n, n))
        if isinstance(i, list):
            if not len(i) == len(j):
                raise ValueError(
                    'mode "' + mode + '": length of index lists must be equal'
                )
            for ind in range(len(i)):
                T[i[ind], j[ind]] = 1
                T[j[ind], i[ind]] = 1
        else:
            T[i, j] = 1
            T[j, i] = 1
    elif mode == "full":
        T = np.ones((n, n))
        for i in range(n):
            T[i, i] = 0
    elif mode == "nearest_int":
        T = np.zeros((n, n))
        for i in range(n - shift):
            if oddeven == "even" and i % 2 == 0:
                continue
            elif oddeven == "odd" and i % 2 == 1:
                continue
            T[i, i + shift] = 1
            T[i + shift, i] = 1
    elif mode == "nearest_int_cyclic":
        T = np.zeros((n, n))
        for i in range(n - 1):
            if oddeven == "even" and i % 2 == 0:
                continue
            elif oddeven == "odd" and i % 2 == 1:
                continue
            T[i, i + 1] = 1
            T[i + 1, i] = 1
        if not oddeven == "even":
            T[0, -1] = 1
            T[-1, 0] = 1
    elif mode == "random":
        T = np.random.rand(n, n)
        for i in range(0, n):
            for j in range(i, n):
                T[i, j] = T[j, i]
    elif mode == "standard" or mode == "Hamming":
        ispowertwo = (n & (n - 1) == 0) and n != 0
        if not ispowertwo:
            raise ValueError('mode "' + mode + '" needs n to be a power of two')
        T = np.zeros((n, n))
        log2n = int(np.log2(n))
        for i in range(n):
            for j in range(n):
                s1 = "{0:b}".format(i).zfill(log2n)
                s2 = "{0:b}".format(j).zfill(log2n)
                if hamming(s1, s2) == d:
                    T[i, j] = 1
    else:
        raise NotImplementedError('mode "' + mode + '" not implemented')

    return T


class PauliStringTP:
    def __init__(self, excludeI=False):
        self.excludeI = excludeI
        self.items = []

    def get_items_PS(self, tp):
        if isinstance(tp, Pauli):
            if not self.excludeI:
                self.items.append(tp)
            elif not tp == I:
                self.items.append(tp)
        else:
            tpL, tpR = tp.args
            if isinstance(tpL, TensorProduct):
                self.get_items_PS(tpL)
            else:
                if not self.excludeI:
                    self.items.append(tpL)
                elif not tpL == I:
                    self.items.append(tpL)
            if isinstance(tpR, TensorProduct):
                self.get_items_PS(tpR)
            else:
                if not self.excludeI:
                    self.items.append(tpR)
                elif not tpR == I:
                    self.items.append(tpR)


def TPoPtoString(item):
    ret = ""
    tps = PauliStringTP()
    tps.get_items_PS(item)
    for p in tps.items:
        if p == 1:
            ret += "I"
        if p == X:
            ret += "X"
        if p == Y:
            ret += "Y"
        if p == Z:
            ret += "Z"
    return ret


def HtoString(H, symbolic=False):
    ret = ""
    if isinstance(H, TensorProduct) or isinstance(
        H, Pauli
    ):  ### go through Pauli string
        ret += TPoPtoString(H)
    else:
        for item in H.args:  ### go through all items of the sum (Pauli strings)
            if isinstance(item, Mul):  ### remove float
                if symbolic:
                    fval, _, item = item.args
                else:
                    if len(item.args) > 2:
                        fval, tmp, item = item.args
                        if not math.isclose(fval, 0, abs_tol=1e-7):
                            raise AssertionError(
                                "Encountered imaginary part that is not close to zero, aborting!",
                                fval,
                                tmp,
                                item,
                            )
                    else:
                        fval, item = item.args
                        if math.isclose(fval, 0, abs_tol=1e-7):
                            item = None
                            print("depug: close to zero", fval, item)
                ret += f"{fval:+.2f}" + " "
            if isinstance(item, TensorProduct) or isinstance(
                item, Pauli
            ):  ### go through Pauli string
                ret += TPoPtoString(item)
            ret += " "
    return ret


def num_Cnot_TPoP(item, symbolic=False):
    tps = PauliStringTP(excludeI=True)
    tps.get_items_PS(item)
    tmp = len(tps.items)
    if tmp == 1:
        sqg = 1
        cnot = 0
    elif tmp > 1:
        sqg = 0
        cnot = 2 * (tmp - 1)
    return sqg, cnot


def num_Cnot(H, symbolic=False):
    sqg = 0
    cnot = 0
    if isinstance(H, TensorProduct) or isinstance(
        H, Pauli
    ):  ### go through Pauli string
        sqg_, cnot_ = num_Cnot_TPoP(H, symbolic=symbolic)
        sqg += sqg_
        cnot += cnot_
    else:
        for item in H.args:  ### go through all items of the sum (Pauli strings)
            if isinstance(item, Mul):  ### remove float
                if symbolic:
                    fval, _, item = item.args
                else:
                    if len(item.args) > 2:
                        fval, tmp, item = item.args
                        if not math.isclose(fval, 0, abs_tol=1e-7):
                            raise AssertionError(
                                "Encountered imaginary part that is not close to zero, aborting!",
                                fval,
                                tmp,
                                item,
                            )
                    else:
                        fval, item = item.args
                        if math.isclose(fval, 0, abs_tol=1e-7):
                            item = None
                            print("depug: close to zero", fval, item)
            if isinstance(item, TensorProduct) or isinstance(
                item, Pauli
            ):  ### go through Pauli string
                sqg_, cnot_ = num_Cnot_TPoP(item, symbolic=symbolic)
                sqg += sqg_
                cnot += cnot_
    return sqg, cnot


def get_g(binstrings):
    n = len(binstrings[0])
    x = binsymbols("x:" + str(n))
    expr = 1
    for bs in binstrings:
        tmp_expr = 0
        for i in range(n):
            if bs[i] == "0":
                tmp_expr += x[i]
            else:
                tmp_expr += (x[i] - 1) ** 2
        expr *= tmp_expr
    return x, expand(expr)


def convert_to_ps(bs1, bs2):
    n = len(bs1)

    for j in range(n):
        if bs1[j] == "1" and bs2[j] == "0":
            tmp = 1 / 2 * (X - 1j * Y)
        elif bs1[j] == "0" and bs2[j] == "1":
            tmp = 1 / 2 * (X + 1j * Y)
        elif bs1[j] == "1" and bs2[j] == "1":
            tmp = 1 / 2 * (I - Z)
        else:  # bs[j]=="0" and bs[j]=="0":
            tmp = 1 / 2 * (I + Z)
        if j == 0:
            pauli_str = tmp
        else:
            pauli_str = TensorProduct(pauli_str, tmp)
    return pauli_str


def get_overlap(binstringsA, binstringsB):
    overlap = []
    mA = len(binstringsA)
    mB = len(binstringsB)
    for i in range(mA):
        for j in range(mB):
            if binstringsA[i] == binstringsB[j]:
                overlap.append(binstringsA[i])
    return overlap


def get_Pauli_string(binstrings, T, symbolic=False):
    m = len(binstrings)

    pauli_str = 0
    if symbolic:
        for i in range(m):
            for j in range(m):
                tmp_ps = convert_to_ps(binstrings[i], binstrings[j])
                pauli_str += T[i, j] * tmp_ps
    else:
        for i in range(m):
            for j in range(m):
                if not math.isclose(T[i, j], 0, abs_tol=1e-7):
                    tmp_ps = convert_to_ps(binstrings[i], binstrings[j])
                    pauli_str += T[i, j] * tmp_ps

    return pauli_str


def get_A1(w, z):
    if z[0] == "0" and w[0] == "0":
        return I + Z
    if (z[0] == "0" and w[0] == "1") or (z[0] == "1" and w[0] == "0"):
        return X
    return I - Z  # z[0]==1 and w[0]==1


def get_B1(w, z):
    if z[0] == "0" and w[0] == "1":
        return -Y
    if z[0] == "1" and w[0] == "0":
        return Y
    return 0.0  # (z[0]==0 and w[0]==0) or (z[0]==1 and w[0]==1)


def get_A_next(x, y, A_n, B_n):
    if x == "0" and y == "0":
        return 0.5 * TensorProduct(A_n, I + Z)
    if x == "0" and y == "1":
        return 0.5 * (TensorProduct(A_n, X) + TensorProduct(B_n, Y))
    if x == "1" and y == "0":
        return 0.5 * (TensorProduct(A_n, X) - TensorProduct(B_n, Y))
    return 0.5 * TensorProduct(A_n, I - Z)  # x==1 and y==1


def get_B_next(x, y, A_n, B_n):
    if x == "0" and y == "0":
        return 0.5 * (TensorProduct(B_n, I + Z))
    if x == "0" and y == "1":
        return 0.5 * (TensorProduct(B_n, X) - TensorProduct(A_n, Y))
    if x == "1" and y == "0":
        return 0.5 * (TensorProduct(B_n, X) + TensorProduct(A_n, Y))
    return 0.5 * (TensorProduct(B_n, I - Z))  # x==1 and y==1


def get_A_n(w, z):
    A_curr = get_A1(w, z)
    B_curr = get_B1(w, z)
    for n in range(1, len(z)):
        x, y = z[n], w[n]
        A_curr, B_curr = get_A_next(x, y, A_curr, B_curr), get_B_next(
            x, y, A_curr, B_curr
        )
    return A_curr


def get_Pauli_string_with_algorithm3(B, T):
    pauli_string = 0
    J = len(B)
    for j in range(J):
        for k in range(j + 1, J):
            if not math.isclose(T[j, k], 0, abs_tol=1e-7):
                w = B[j]
                z = B[k]
                A_n = get_A_n(w, z)  # recursively compute A_n
                pauli_string += A_n * T[j, k]
    return pauli_string


def simplifyH(H):
    for i in range(10):
        H = H.expand(tensorproduct=True)
    H = evaluate_pauli_product(H)
    return H


def get_H(stringlist, T, simplify=True, symbolic=False, verbose=False):
    H = get_Pauli_string(stringlist, T, symbolic=symbolic)
    if simplify:
        H = simplifyH(H)
    if verbose:
        print("#sqg, #cnots=", num_Cnot(H, symbolic=symbolic))
    return H


def matrix_to_sympy(H):
    """Decompose matrix H to sympy pauli string"""
    array = [i.split("*") for i in decompose(H).replace(" ", "").split("+")[1:]]
    pauli_str_sympy = 0
    pauli_map = {"I": 1, "X": Pauli(1), "Y": Pauli(2), "Z": Pauli(3)}
    for number, pauli_str in array:
        if len(pauli_str) == 1:
            tp = pauli_map[pauli_str]
        else:
            tp = TensorProduct(pauli_map[pauli_str[0]], pauli_map[pauli_str[1]])
            for i in range(2, len(pauli_str)):
                tp = TensorProduct(tp, pauli_map[pauli_str[i]])
        pauli_str_sympy += float(number) * tp
    return pauli_str_sympy


class Circuit_maker:
    """class for creating circuit from sympy expression"""

    def __init__(self, H):
        self.array = self.__create_array(H)

    def add_gates(self, circuit, parameter):
        """iterate over self.array and add gates to circuit,
        parameter is the rotation angle used in the rz gate"""
        for row in self.array:
            if row[0] == "h":
                circuit.h(row[1])
            elif row[0] == "cx":
                circuit.cx(row[1], row[2])
            elif row[0] == "rz":
                circuit.rz(row[1] * parameter, row[2])
            elif row[0] == "s":
                circuit.s(row[1])
            elif row[0] == "sdg":
                circuit.sdg(row[1])

    def __create_array(self, H):
        """create self.array which holds instructions of where gates should be placed,
        corresponding to equation 33 in Fuchs et al. (arXiv:2203.06095).
        H is a sympy expression and could be for example the mixing operator, H_m"""
        array = []
        if H != 0:
            commuting_terms = Add.make_args(H)
            for expr in commuting_terms:
                # remove artificially introduced symbols from H, which was introduced to keep
                # commuting terms separated. Example: expr = c(4*II + XX) -> H_args = [4*II, XX].
                H_args = Add.make_args(
                    list(
                        filter(
                            lambda x: not isinstance(x, Symbol),
                            list(Mul.make_args(expr)),
                        )
                    )[0]
                )
                if len(H_args) == 1:  # if only one term in H
                    H_args = [H_args[0]]

                for pauli_string in H_args:
                    # remove all symbols from expression, and split pauli string to (float, Symbol, TensorProduct)
                    mul_args = list(
                        filter(
                            lambda x: not isinstance(x, Symbol),
                            list(Mul.make_args(pauli_string)),
                        )
                    )
                    angle, tensor_prod = mul_args

                    # split tensor product expression on each tensor product
                    operators = self.__get_operators_from_tensor_product(tensor_prod)

                    # add gates to array
                    self.__add_U(array, operators)
                    self.__add_middle_part(array, operators, angle)
                    self.__add_U_dagger(array, operators)
        return array

    def __get_operators_from_tensor_product(self, tensor_product):
        """split tensor product expr on each tensor product"""
        operators = []
        while not isinstance(tensor_product, (Pauli, sympyOne)):
            tensor_product, operator = tensor_product.args
            operators.insert(0, operator)
        operators.insert(0, tensor_product)
        return operators

    def __add_U(self, array, operators):
        """add gates corresponding to block U in eq. 33"""
        for i, operator in enumerate(operators):
            if operator == Pauli(1):  # X
                array.append(("h", i))
            elif operator == Pauli(2):  # Y
                array.append(("s", i))
                array.append(("h", i))

    def __add_middle_part(self, array, operators, angle):
        """add cnots and rz gate corresponding to the middle part in eq. 33"""
        pairs = self.__get_cx_pairs(operators)

        # if only one operator in operators, add rz gate on correct index (do not add cx gates)
        if len(pairs) == 0:
            for i, operator in enumerate(operators):
                if isinstance(operator, Pauli):
                    array.append(("rz", -2.0 * float(angle), i))
                    break
        else:
            for pair in pairs:
                array.append(("cx", pair[0], pair[1]))
            array.append(("rz", -2.0 * float(angle), pairs[-1][1]))
            for pair in pairs[::-1]:
                array.append(("cx", pair[0], pair[1]))

    def __add_U_dagger(self, array, operators):
        """add gates corresponding to block dagger(U) in eq. 33"""
        for i, operator in enumerate(operators):
            if operator == Pauli(1):  # X
                array.append(("h", i))
            elif operator == Pauli(2):  # Y
                array.append(("h", i))
                array.append(("sdg", i))

    def __get_cx_pairs(self, operators):
        """return list of indices of where cx gates should be placed"""
        pairs = []
        j = None
        for i, operator in enumerate(operators):
            if i != 0 and isinstance(operator, Pauli) and j != None:
                pairs.append((j, i))
            if isinstance(operator, Pauli):
                j = i
        return pairs



import math

import numpy as np

from .qubo_problem import QUBO


class PortfolioOptimization(QUBO):
    def __init__(self, risk, budget, cov_matrix, exp_return, penalty=0) -> None:
        self.risk = risk
        self.budget = budget
        self.cov_matrix = cov_matrix
        self.exp_return = exp_return
        self.penalty = penalty
        self.N_qubits = len(self.exp_return)

        # Reformulated as a QUBO
        # min x^T Q x + c^T x + b
        # Writing Q as lower triangular matrix since it otherwise is symmetric
        Q = self.risk * np.tril(
            self.cov_matrix + np.tril(self.cov_matrix, k=-1)
        ) + self.penalty * (
            np.eye(self.N_qubits)
            + 2 * np.tril(np.ones((self.N_qubits, self.N_qubits)), k=-1)
        )
        c = -self.exp_return - (
            2 * self.penalty * self.budget * np.ones_like(self.exp_return)
        )
        b = self.penalty * self.budget * self.budget

        super().__init__(Q=Q, c=c, b=b)

    def cost_nonQUBO(self, string, penalize=True):
        # risk = self.params.get("risk")
        # budget = self.params.get("budget")
        # cov_matrix = self.params.get("cov_matrix")
        # exp_return = self.params.get("exp_return")
        # penalty = self.params.get("penalty", 0.0)

        x = np.array(list(map(int, string)))
        cost = risk * (x.T @ cov_matrix @ x) - exp_return.T @ x
        if penalize:
            cost += penalty * (x.sum() - budget) ** 2

        return -cost

    def isFeasible(self, string):
        x = self.__str2np(string)
        constraint = np.sum(x) - self.budget
        return math.isclose(constraint, 0, abs_tol=1e-7)

    def __str2np(self, s):
        x = np.array(list(map(int, s)))
        assert len(x) == len(self.exp_return), (
            "bitstring  "
            + s
            + " of wrong size. Expected "
            + str(len(self.exp_return))
            + " but got "
            + str(len(x))
        )
        return x



from qiskit import QuantumCircuit, QuantumRegister
from qiskit.circuit import Parameter

from abc import abstractmethod

from .base_mixer import Mixer


class Constrained(Mixer):
    def __init__(self) -> None:
        self.B = []
        self.best_mixer_terms = []
        self.circuit = None
        # self.reduced = self.params.get("reduced", True)

    # def set_initial_state(self, circuit, qubit_register):
    #    # set to ground state of mixer hamiltonian??
    #    if not self.B:
    #        self.compute_feasible_subspace()
    #        # initial state
    #    ampl_vec = np.zeros(2 ** len(self.B[0]))
    #    ampl = 1 / np.sqrt(len(self.B))
    #    for state in self.B:
    #        ampl_vec[int(state, 2)] = ampl

    #    circuit.initialize(ampl_vec, qubit_register)

    def create_circuit(self):
        if not self.B:
            self.compute_feasible_subspace()

        m = Mixer(self.B, sort=True)
        m.compute_commuting_pairs()
        m.compute_family_of_graphs()
        m.get_best_mixer_commuting_graphs(reduced=self.reduced)
        (
            self.circuit,
            self.best_mixer_terms,
            self.logical_X_operators,
        ) = m.compute_parametrized_circuit(self.reduced)

    @abstractmethod
    def compute_feasible_subspace(self):
        pass

    def set_feasible_subspace(self, space):
        self.B = space

    @abstractmethod
    def isFeasible(self, string):
        pass



import structlog

LOG = structlog.get_logger(file=__name__)

import numpy as np

from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister, execute, Aer
from qiskit.circuit import Parameter
from qiskit.primitives import Sampler
from qiskit.algorithms.optimizers import COBYLA

from qaoa.initialstates import InitialState
from qaoa.mixers import Mixer
from qaoa.problems import Problem
from qaoa.util import Statistic


class OptResult:
    """
    Class for results of optimization at depth p
    """

    def __init__(self, depth):
        self.depth = depth

        self.angles = []
        self.Exp = []
        self.Var = []
        self.MaxCost = []
        self.MinCost = []
        self.shots = []

        self.index_Exp_min = -1

    def add_iteration(self, angles, stat, shots):
        self.angles.append(angles)
        self.Exp.append(-stat.get_CVaR())
        self.Var.append(stat.get_Variance())
        self.MaxCost.append(-stat.get_max())
        self.MinCost.append(-stat.get_min())
        self.shots.append(shots)

    def compute_best_index(self):
        self.index_Exp_min = self.Exp.index(min(self.Exp))

    def get_best_Exp(self):
        return self.Exp[self.index_Exp_min]

    def get_best_Var(self):
        return self.Var[self.index_Exp_min]

    def get_best_angles(self):
        return self.angles[self.index_Exp_min]

    def num_fval(self):
        return len(self.Exp)

    def num_shots(self):
        return sum(self.shots)


class QAOA:
    """Main class"""

    def __init__(
        self,
        problem,
        mixer,
        initialstate,
        backend=Aer.get_backend("qasm_simulator"),
        noisemodel=None,
        optimizer=[COBYLA, {}],  # optimizer, options
        precision=None,
        shots=1024,
        alpha=1,
    ) -> None:
        """
        A QAO-Ansatz consist of these parts:

        :problem of Basetype Problem,
        implementing the phase circuit and the cost.

        :mixer of Basetype Mixer,
        specifying the mixer circuit.

        :initialstate of Basetype InitialState,
        specifying the initial state.

        :backend: backend
        :noisemodel: noisemodel
        :optmizer: optimizer
        :precision: precision to reach for expectation value based on error=variance/sqrt(shots)
        :shots: if precision=None, the number of samples taken
                      if precision!=None, the minimum number of samples taken
        :alpha: used for CVar
        """

        assert issubclass(type(problem), Problem)
        assert issubclass(type(mixer), Mixer)
        assert issubclass(type(initialstate), InitialState)

        self.problem = problem
        self.mixer = mixer
        self.initialstate = initialstate
        self.initialstate.setNumQubits(self.problem.N_qubits)
        self.mixer.setNumQubits(self.problem.N_qubits)

        self.parameterized_circuit = None
        self.parameterized_circuit_depth = 0

        self.backend = backend
        self.optimizer = optimizer
        self.noisemodel = noisemodel
        self.shots = shots
        self.precision = precision
        self.stat = Statistic(alpha=alpha)

        self.usebarrier = False
        self.isQNSPSA = False

        self.current_depth = 0  # depth at which local optimization has been done

        self.gamma_params = None
        self.beta_params = None

        self.Exp_sampled_p1 = None
        self.landscape_p1_angles = {}
        self.Var_sampled_p1 = None
        self.MaxCost_sampled_p1 = None
        self.MinCost_sampled_p1 = None

        self.optimization_results = {}

    def exp_landscape(self):
        ### at depth p = 1
        return self.Exp_sampled_p1

    def var_landscape(self):
        ### at depth p = 1
        return self.Var_sampled_p1

    def get_Exp(self, depth=None):
        if not depth:
            ret = []
            for i in range(1, self.current_depth + 1):
                ret.append(self.optimization_results[i].get_best_Exp())
            return ret
        if depth > self.current_depth + 1:
            raise ValueError
        return self.optimization_results[depth].get_best_Exp()

    def get_Var(self, depth):
        if depth > self.current_depth + 1:
            raise ValueError
        return self.optimization_results[depth].get_best_Var()

    def get_beta(self, depth):
        if depth > self.current_depth + 1:
            raise ValueError
        return self.optimization_results[depth].get_best_angles()[1::2]

    def get_gamma(self, depth):
        if depth > self.current_depth + 1:
            raise ValueError
        return self.optimization_results[depth].get_best_angles()[::2]

    def createParameterizedCircuit(self, depth):
        if self.parameterized_circuit_depth == depth:
            LOG.info(
                "Circuit is already of depth " + str(self.parameterized_circuit_depth)
            )
            return

        self.problem.create_circuit()
        self.mixer.create_circuit()
        self.initialstate.create_circuit()

        q = QuantumRegister(self.problem.N_qubits)
        c = ClassicalRegister(self.problem.N_qubits)
        self.parameterized_circuit = QuantumCircuit(q, c)

        self.gamma_params = [None] * depth
        self.beta_params = [None] * depth

        self.parameterized_circuit.compose(self.initialstate.circuit, inplace=True)

        if self.usebarrier:
            self.circuit.barrier()

        for d in range(depth):
            self.gamma_params[d] = Parameter("gamma_" + str(d))
            tmp_circuit = self.problem.circuit.assign_parameters(
                {self.problem.circuit.parameters[0]: self.gamma_params[d]},
                inplace=False,
            )
            self.parameterized_circuit.compose(tmp_circuit, inplace=True)

            if self.usebarrier:
                self.circuit.barrier()

            self.beta_params[d] = Parameter("beta_" + str(d))
            tmp_circuit = self.mixer.circuit.assign_parameters(
                {self.mixer.circuit.parameters[0]: self.beta_params[d]},
                inplace=False,
            )
            self.parameterized_circuit.compose(tmp_circuit, inplace=True)

            if self.usebarrier:
                self.circuit.barrier()

        self.parameterized_circuit.barrier()
        self.parameterized_circuit.measure(q, c)
        self.parametrized_circuit_depth = depth

    def sample_cost_landscape(
        self,
        angles={"gamma": [0, 2 * np.pi, 20], "beta": [0, 2 * np.pi, 20]},
    ):
        self.landscape_p1_angles = angles
        logger = LOG.bind(func=self.sample_cost_landscape.__name__)
        logger.info("Calculating energy landscape for depth p=1...")

        depth = 1

        tmp = angles["gamma"]
        self.gamma_grid = np.linspace(tmp[0], tmp[1], tmp[2])
        tmp = angles["beta"]
        self.beta_grid = np.linspace(tmp[0], tmp[1], tmp[2])

        if self.backend.configuration().local:
            print(depth, self.current_depth)
            self.createParameterizedCircuit(depth)
            gamma = [None] * angles["beta"][2] * angles["gamma"][2]
            beta = [None] * angles["beta"][2] * angles["gamma"][2]

            counter = 0
            for b in range(angles["beta"][2]):
                for g in range(angles["gamma"][2]):
                    gamma[counter] = self.gamma_grid[g]
                    beta[counter] = self.beta_grid[b]
                    counter += 1

            parameters = {self.gamma_params[0]: gamma, self.beta_params[0]: beta}

            logger.info("Executing sample_cost_landscape")
            logger.info(f"parameters: {len(parameters)}")
            job = execute(
                self.parameterized_circuit,
                self.backend,
                shots=self.shots,
                parameter_binds=[parameters],
                optimization_level=0,
            )
            logger.info("Done execute")
            self.measurementStatistics(job)
            logger.info("Done measurement")

        else:
            raise NotImplementedError

        logger.info("Calculating Energy landscape done")

    def measurementStatistics(self, job):
        """
        implements a function for expectation value and variance

        :param job: job instance derived from BaseJob
        :return: expectation and variance, if the job is a list
        """
        jres = job.result()
        counts_list = jres.get_counts()
        if isinstance(counts_list, list):
            expectations = []
            variances = []
            maxcosts = []
            mincosts = []
            for i, counts in enumerate(counts_list):
                self.stat.reset()
                for string in counts:
                    # qiskit binary strings use little endian encoding, but our cost function expects big endian encoding. Therefore, we reverse the order
                    cost = self.problem.cost(string[::-1])
                    self.stat.add_sample(cost, counts[string])
                expectations.append(self.stat.get_CVaR())
                variances.append(self.stat.get_Variance())
                maxcosts.append(self.stat.get_max())
                mincosts.append(self.stat.get_min())
            angles = self.landscape_p1_angles
            self.Exp_sampled_p1 = -np.array(expectations).reshape(
                angles["beta"][2], angles["gamma"][2]
            )
            self.Var_sampled_p1 = np.array(variances).reshape(
                angles["beta"][2], angles["gamma"][2]
            )
            self.MaxCost_sampled_p1 = -np.array(maxcosts).reshape(
                angles["beta"][2], angles["gamma"][2]
            )
            self.MinCost_sampled_p1 = -np.array(mincosts).reshape(
                angles["beta"][2], angles["gamma"][2]
            )
        else:
            for string in counts_list:
                # qiskit binary strings use little endian encoding, but our cost function expects big endian encoding. Therefore, we reverse the order
                cost = self.problem.cost(string[::-1])
                self.stat.add_sample(cost, counts_list[string])

    def optimize(self, depth):
        ## run local optimization by iteratively increasing the depth until depth p is reached
        while self.current_depth < depth:
            if self.current_depth == 0:
                if self.Exp_sampled_p1 is None:
                    self.sample_cost_landscape()
                ind_Emin = np.unravel_index(
                    np.argmin(self.Exp_sampled_p1, axis=None), self.Exp_sampled_p1.shape
                )
                angles0 = np.array(
                    (self.gamma_grid[ind_Emin[1]], self.beta_grid[ind_Emin[0]])
                )
            else:
                gamma = self.optimization_results[self.current_depth].get_best_angles()[
                    ::2
                ]
                beta = self.optimization_results[self.current_depth].get_best_angles()[
                    1::2
                ]

                gamma_interp = self.interp(gamma)
                beta_interp = self.interp(beta)
                angles0 = np.zeros(2 * (self.current_depth + 1))
                angles0[::2] = gamma_interp
                angles0[1::2] = beta_interp

            self.optimization_results[self.current_depth + 1] = OptResult(
                self.current_depth + 1
            )
            # Create parameterized circuit at new depth
            self.createParameterizedCircuit(int(len(angles0) / 2))

            res = self.local_opt(angles0)

            self.optimization_results[self.current_depth + 1].compute_best_index()

            LOG.info(
                f"cost(depth { self.current_depth + 1} = {res.fun}",
                func=self.optimize.__name__,
            )

            self.current_depth += 1

    def local_opt(self, angles0):
        """

        :param angles0: initial guess
        """

        try:
            opt = self.optimizer[0](**self.optimizer[1])
        except TypeError as e:  ### QNSPSA needs fidelity
            self.isQNSPSA = True
            self.optimizer[1]["fidelity"] = self.optimizer[0].get_fidelity(
                self.parameterized_circuit, sampler=Sampler()
            )
            opt = self.optimizer[0](**self.optimizer[1])
        res = opt.minimize(self.loss, x0=angles0)
        if self.isQNSPSA:
            self.optimizer[1].pop("fidelity")
        return res

    def loss(self, angles):
        """
        loss function
        :return: an instance of the qiskit class QuantumCircuit
        """

        circuit = None
        n_target = self.shots
        self.stat.reset()
        shots_taken = 0
        shots = self.shots

        for i in range(3):  # this loop is used only used if precision is set
            if self.backend.configuration().local:
                params = self.getParametersToBind(
                    angles, self.parametrized_circuit_depth, asList=True
                )
                job = execute(
                    self.parameterized_circuit,
                    backend=self.backend,
                    noise_model=self.noisemodel,
                    shots=shots,
                    parameter_binds=[params],
                    optimization_level=0,
                )
            else:
                raise NotImplementedError
            shots_taken += shots
            self.measurementStatistics(job)
            if self.precision is None:
                break
            else:
                v = self.stat.get_Variance()
                shots = int((np.sqrt(v) / self.precision) ** 2) - shots_taken
                if shots <= 0:
                    break

        self.optimization_results[self.current_depth + 1].add_iteration(
            angles.copy(), self.stat, shots_taken
        )

        return -self.stat.get_CVaR()

    def getParametersToBind(self, angles, depth, asList=False):
        """
        Utility function to structure the parameterized parameter values
        so that they can be applied/bound to the parameterized circuit.

        :param angles: gamma and beta values
        :param depth: circuit depth
        :asList: Boolean that specify if the values in the dict should be a list or not
        :return: A dict containing parameters as keys and parameter values as values
        """
        assert len(angles) == 2 * depth

        params = {}
        for d in range(depth):
            if asList:
                params[self.gamma_params[d]] = [angles[2 * d + 0]]
                params[self.beta_params[d]] = [angles[2 * d + 1]]
            else:
                params[self.gamma_params[d]] = angles[2 * d + 0]
                params[self.beta_params[d]] = angles[2 * d + 1]
        return params

    def interp(self, angles):
        """
        INTERP heuristic/linear interpolation for initial parameters
        when going from depth p to p+1 (https://doi.org/10.1103/PhysRevX.10.021067)
        E.g. [0., 2., 3., 6., 11., 0.] becomes [2., 2.75, 4.5, 7.25, 11.]

        :param angles: angles for depth p
        :return: linear interpolation of angles for depth p+1
        """
        depth = len(angles)
        tmp = np.zeros(len(angles) + 2)
        tmp[1:-1] = angles.copy()
        w = np.arange(0, depth + 1)
        return w / depth * tmp[:-1] + (depth - w) / depth * tmp[1:]

    def hist(self, angles):
        depth = int(len(angles) / 2)
        self.createParameterizedCircuit(depth)

        params = self.getParametersToBind(angles, depth, asList=True)
        if self.backend.configuration().local:
            job = execute(
                self.parameterized_circuit,
                self.backend,
                shots=self.shots,
                parameter_binds=[params],
                optimization_level=0,
            )
        else:
            job = start_or_retrieve_job(
                "hist", self.backend, circ, options={"shots": self.shots}
            )
        return job.result().get_counts()

    def random_init(self, gamma_bounds, beta_bounds, depth):
        """
        Enforces the bounds of gamma and beta based on the graph type.
        :param gamma_bounds: Parameter bound tuple (min,max) for gamma
        :param beta_bounds: Parameter bound tuple (min,max) for beta
        :return: np.array on the form (gamma_1, beta_1, gamma_2, ...., gamma_d, beta_d)
        """
        gamma_list = np.random.uniform(gamma_bounds[0], gamma_bounds[1], size=depth)
        beta_list = np.random.uniform(beta_bounds[0], beta_bounds[1], size=depth)
        initial = np.empty((gamma_list.size + beta_list.size,), dtype=gamma_list.dtype)
        initial[0::2] = gamma_list
        initial[1::2] = beta_list
        return initial



import math

import numpy as np

from .base_problem import Problem
from qiskit import QuantumCircuit, QuantumRegister

from qiskit.circuit import Parameter


class ExactCover(Problem):
    def __init__(
        self,
        FR,
        CR=None,
        mu=1,
    ) -> None:
        self.FR = FR
        self.CR = CR
        self.mu = 1

        fN = FR.shape[0]  ### number of flights
        rN = FR.shape[1]  ### number of routes

        self.N_qubits = rN

    def cost(self, string):
        x = np.array(list(map(int, string)))
        c_e = self.__exactCover(x)

        if self.CR is None:
            return -c_e
        else:
            return -(self.CR @ x + self.mu * c_e)

    def create_circuit(self):
        """
        Creates parameterized circuit corresponding to the cost function
        """
        q = QuantumRegister(self.N_qubits)
        self.circuit = QuantumCircuit(q)
        cost_param = Parameter("x_gamma")

        F, R = np.shape(self.FR)

        ### cost Hamiltonian
        for r in range(R):
            hr = self.mu * 0.5 * self.FR[:, r] @ (np.sum(self.FR, axis=1) - 2)
            if not self.CR is None:
                hr += 0.5 * self.CR[r]

            if not math.isclose(hr, 0, abs_tol=1e-7):
                self.circuit.rz(cost_param * hr, q[r])

            for r_ in range(r + 1, R):
                Jrr_ = self.mu * 0.5 * self.FR[:, r] @ self.FR[:, r_]

                if not math.isclose(Jrr_, 0, abs_tol=1e-7):
                    self.circuit.cx(q[r], q[r_])
                    self.circuit.rz(cost_param * Jrr_, q[r_])
                    self.circuit.cx(q[r], q[r_])

    def isFeasible(self, string):
        x = np.array(list(map(int, string)))
        c_e = self.__exactCover(x)
        return math.isclose(c_e, 0, abs_tol=1e-7)

    def __exactCover(self, x):
        return np.sum((1 - (self.FR @ x)) ** 2)



import numpy as np
from qiskit import QuantumRegister, QuantumCircuit
from qiskit.circuit.library import RYGate


def getRYi(n):
    """
    Returns gate (i) from section 2.2.

    Args:
        n (int): The integer parameter for gate (i).

    Returns:
        QuantumCircuit: Quantum circuit representing gate (i).
    """

    qc = QuantumCircuit(2)

    qc.cnot(0, 1)
    theta = 2 * np.arccos(np.sqrt(1 / n))
    ry = RYGate(theta).control(ctrl_state="1")
    qc.append(ry, [1, 0])
    qc.cnot(0, 1)

    return qc


def getRYii(l, n):
    """
    Returns gate (ii)_l from section 2.2.

    Args:
        l (int): The integer parameter for gate (ii)_l.
        n (int): The integer parameter for gate (ii)_l.

    Returns:
        QuantumCircuit: Quantum circuit representing gate (ii)_l.
    """

    qc = QuantumCircuit(3)

    qc.cnot(0, 2)
    theta = 2 * np.arccos(np.sqrt(l / n))
    ry = RYGate(theta).control(num_ctrl_qubits=2, ctrl_state="11")
    qc.append(ry, [2, 1, 0])
    qc.cnot(0, 2)

    return qc


def getSCS(n, k):
    """
    Returns SCS_{n,k} gate from definition 3.

    Args:
        n (int): The integer parameter for SCS_{n,k}.
        k (int): The integer parameter for SCS_{n,k}.

    Returns:
        QuantumCircuit: Quantum circuit representing SCS_{n,k}.
    """

    qc = QuantumCircuit(k + 1)

    qc.append(getRYi(n), [k - 1, k])
    for l in range(2, k + 1):
        qc.append(getRYii(l, n), [k - l, k - l + 1, k])

    return qc


def getBlock1(n, k, l):
    """
    Returns the first block in Lemma 2.

    Args:
        n (int): The integer parameter for the quantum register size.
        k (int): The integer parameter for the Hamming weight.
        l (int): The integer parameter for the block.

    Returns:
        QuantumCircuit: Quantum circuit representing the first block in Lemma 2.
    """

    qr = QuantumRegister(n)
    qc = QuantumCircuit(qr)

    first = l - k - 1
    last = n - l

    index = list(range(n))

    if first != 0:
        index = index[first:]
        qc.i(qr[:first])

    if last != 0:
        index = index[:-last]
        qc.append(getSCS(l, k), index)
        qc.i(qr[-last:])
    else:
        qc.append(getSCS(l, k), index)

    return qc


def getBlock2(n, k, l):
    """
    Returns the second block from Lemma 2.

    Args:
        n (int): The integer parameter for the quantum register size.
        k (int): The integer parameter for the Hamming weight.
        l (int): The integer parameter for the block.

    Returns:
        QuantumCircuit: Quantum circuit representing the second block in Lemma 2.
    """

    qr = QuantumRegister(n)
    qc = QuantumCircuit(qr)

    last = n - l
    index = list(range(n))

    if last != 0:
        index = index[:-last]
        qc.append(getSCS(l, l - 1), index)
        qc.i(qr[-last:])
    else:
        qc.append(getSCS(l, l - 1), index)

    return qc


def DickeState(n, k, barrier=False):
    """
    Circuit to prepare Dicke states, following the algorithm from https://arxiv.org/pdf/1904.07358.pdf.

    Args:
        n (int): The number of qubits in the quantum register.
        k (int): The Hamming weight of the Dicke states.
        barrier (bool, optional): Whether to insert barriers between gates for clarity (default is False).

    Returns:
        QuantumCircuit: Quantum circuit to prepare Dicke states.
    """

    qr = QuantumRegister(n)
    qc = QuantumCircuit(qr)

    qc.x(qr[-k:])

    for l in range(k + 1, n + 1)[::-1]:
        qc.append(getBlock1(n, k, l), range(n))

    for l in range(2, k + 1)[::-1]:
        qc.append(getBlock2(n, k, l), range(n))

    if barrier:
        qc.barrier()

    return qc
